import jax.numpy as jnp
from jax import grad, jit, vmap, lax
from functools import partial

from typing import NamedTuple, Tuple

class MotorParams(NamedTuple):
    kv: float
    kt: float
    resistance: float
    friction_coeff: float
    iron_coeff: float


class BatteryParams(NamedTuple):
    shape: Tuple[int, int] # (series,parallel) array of batteries
    resistance: float # ohms
    initial_energy: float # joules

class CarParams(NamedTuple):
    """ Physical Data for Solar Car Parameters """
    mass: float = 800 # kg
    frontal_area: float = 1.3 # m^2
    drag_coeff: float = 0.018 # drag coefficient, dimensionless
    rolling_coeff: float = 0.002 # rolling resistance.
    moter_eff: float = 0.93 # 0 < x < 1 scaling factor
    solar_area: float = 5.0 # m^2, typically 5.0
    solar_eff: float = 0.20 # 0 < x < 1, typically ~.25
    n_motors: int = 2 # how many motors we have.
    motor: MotorParams = MotorParams(8.43, 1.1, 100.0, 0.001, 0.001) # mitsuba m2090 estimate
    battery: BatteryParams = BatteryParams((36,19), 0.0126, 66.6e3) # freebasing 50s pack.



def DefaultCar() -> CarParams:
    """ Creates a basic car """
    return CarParams(1000, 1.3, 0.18, 0.002, 0.85, 5.0, 0.23)



# some physics equations using jax


@jit
def normal_force(mass, theta):
    return mass * 9.8 * jnp.cos(theta)

@jit
def downslope_force(mass, theta):
    return mass * 9.8 * jnp.sin(theta)

@partial(jit, static_argnames=['crr'])
def rolling_force(mass, theta, crr):
    return normal_force(mass, theta) * crr

@partial(jit, static_argnames=['area', 'cd', 'rho'])
def drag_force(u, area, cd, rho):
    return 0.5 * rho * jnp.pow(u, 2) * cd * area

# we can use those forces above to determine what forces we have to overcome. Sum(F)=0

@partial(jit, static_argnums=(3,4,5,6,7,))
def bldc_power_draw(torque, velocity, resistance=0.1, kt=0.1, 
                   Cf=0.01, iron_loss_coeff=0.005):
    """
    Approximates power draw of a BLDC motor outputting a torque at a given velocity
    
    Args:
        torq: Applied force in Newton/meters
        velocity: Angular velocity in rad/s
        resistance: Motor phase resistance (ohms)
        kt: Torque constant (N⋅m/A)
        friction_coeff: Mechanical friction coefficient
        iron_loss_coeff: Iron loss coefficient (core losses)
    
    Returns:
        Total electrical power draw in Watts
    """
    
    # Current required for torque (simplified relationship)
    current = torque / kt
    
    # Copper losses (I²R)
    copper_losses = resistance * current**2 
    # Mechanical friction losses
    friction_losses = Cf * velocity**2 
    # Iron losses (simplified model - primarily dependent on speed)
    iron_losses = iron_loss_coeff * velocity**2 
    # Mechanical power output
    mechanical_power = torque * velocity
    
    # Total electrical power input
    total_power = mechanical_power + copper_losses + friction_losses + iron_losses
    
    return total_power

@partial(jit, static_argnames=['resistance', 'kt', 'kv', 'vmax', 'Cf'])
def bldc_torque(velocity, current_limit, resistance, kt, kv, vmax, Cf):

    bemf = velocity / kv
    v_avail = jnp.clip(vmax - bemf, 0.0, vmax)
    current = jnp.clip(v_avail / resistance, 0.0, current_limit)

    torque = kt * current
    friction_torque = Cf * velocity
    net_torque = jnp.maximum(torque - friction_torque, 0.0)
    stall_torque = kt * current_limit
    return jnp.where(velocity < 0.01, stall_torque, net_torque)

@partial(jit, static_argnums=(2,3,))
def battery_powerloss(current,cell_r, battery_shape: Tuple[int,int]):
    r_array = jnp.full(battery_shape, cell_r)
    branch_current = current / battery_shape[1]
    I_array = jnp.full(battery_shape, branch_current)
    cell_Ploss = jnp.square(I_array) * r_array
    return jnp.sum(cell_Ploss)



def forward(state, timestep, control, params: CarParams):
    # state is (position, time, velocity, energy)
    # control is -1 to 1 (motor max current percent.)
    # timestep is >0 time to advance
    # params is the params dictionary.
    # returns the next state with (position', time + timestep, velocity', energy')
    # TODO: terrain, weather, solar
    
    # determine the forces acting on the car.
    dragf = drag_force(state[2], params.frontal_area, params.drag_coeff, 1.184)
    rollf = rolling_force(params.mass, 0, params.rolling_coeff)
    hillforce = downslope_force(params.mass, 0)

    pass
